local IsInit
local SpellIdToButton
local ActionToId
local ActionToButton
local ButtonToAction
local DebugSpellMode
local AverageMode
local playerGuid
local EventHandlersTable
local EventFrame = CreateFrame("Frame")


local p_timestamp 		= 1
local p_event			= 2
local p_hideCaster      = 3
local p_sourceGUID      = 4
local p_sourceName      = 5
local p_sourceFlags     = 6
local p_sourceRaidFlags = 7
local p_destGUID        = 8
local p_destName        = 9
local p_destFlags       = 10
local p_destRaidFlags   = 11

--offset 12-13(enviroinmental)
local p_spell_offset	= 11
local p_spell_id		= 1
local p_spell_name		= 2
local p_spell_school	= 3

local p_effect_offset	= 14
local p_effect_amount	= 1
local p_effect_overkill	= 2
local p_effect_school	= 3

local p_buff_offset		= 15

local spell_effects = {}

local function GetFormattedValue(v)
	if v < 1000 then
		return string.format("%d", v)
	elseif v < 10000 then
		return string.format("%.2fk", v / 1000)
	elseif v < 100000 then
		return string.format("%.1fk", v / 1000)
	elseif v < 1000000 then
		return string.format("%dk", v / 1000)
	else
		return string.format("%.2fm", v / 1000000)
	end
end

local PER_PAGE		= 12
local PAGE1_FROM 	= 73
local PAGE1_TO		= PAGE1_FROM + PER_PAGE - 1
local PAGE2_FROM 	= 13
local PAGE2_TO		= PAGE2_FROM + PER_PAGE - 1

local AVERAGE_COUNT_MAX = 10

local function SetButtonText(spellId, text)
	--print("SetButtonText", spellId, text)
	--get proper button
	local page = GetActionBarPage()
	local id = spellId
	if type(SpellIdToButton[id]) == "table" then
		for k, btn in pairs(SpellIdToButton[id]) do
			if btn then
				--print("button found", btn:GetName())
				--print(id, btn:GetName(), btn.action, page)
				if btn.action >= PAGE1_FROM and btn.action <= PAGE1_TO then
					if page == 1 then
						btn.damagetext:SetText(text)
						btn.damagetext:Show()
					else
						btn.damagetext:Hide()
					end
				elseif btn.action >= PAGE2_FROM and btn.action <= PAGE2_TO then
					if page == 2 then
						btn.damagetext:SetText(text)
						btn.damagetext:Show()
					else
						btn.damagetext:Hide()
					end
				else
					btn.damagetext:SetText(text)
					btn.damagetext:Show()
				end
			end
		end
	end
end

local function ResetSpell(id)
	--reset amounts by spell on next cast
	if spell_effects[id] then
		spell_effects[id].damage.current = 0
		spell_effects[id].heal.current = 0
	end
end

local function ShowEffect(id)
	--print("ShowEffect", id)
	if spell_effects[id] then
		--calculate accumulated and show
		if spell_effects[id].damage.current > 0 and spell_effects[id].damage.prev > 0 then
			SetButtonText(id, GetFormattedValue(spell_effects[id].damage.prev))
		elseif spell_effects[id].heal.current > 0 and spell_effects[id].heal.prev > 0 then
			SetButtonText(id, GetFormattedValue(spell_effects[id].heal.prev))
		end
	end
end

local function InitAndShowSpell(id)
	--print("InitAndShowSpell", id)
	spell_effects[id] = spell_effects[id] or {}
	spell_effects[id].damage = spell_effects[id].damage or {["current"] = 0, ["prev"] = 0, ["count"] = 0}
	spell_effects[id].heal = spell_effects[id].heal or {["current"] = 0, ["prev"] = 0, ["count"] = 0}
	--update amounts by spell
	if spell_effects[id].damage.current > 0 then
		local t = spell_effects[id].damage
		if AverageMode then
			if t.count < AVERAGE_COUNT_MAX then
				t.count = t.count + 1
			end
			if t.prev == 0 then
				t.prev = t.current
			else
				t.prev = (t.prev * t.count + t.current) / (t.count + 1)
				--print(t.prev, t.current, t.count)
			end
		else
			t.prev = t.current
		end
	end
	if spell_effects[id].heal.current > 0 then
		local t = spell_effects[id].heal
		if AverageMode then
			if t.count < AVERAGE_COUNT_MAX then
				t.count = t.count + 1
			end
			if t.prev == 0 then
				t.prev = t.current
			else
				t.prev = (t.prev * t.count + t.current) / (t.count + 1)
			end
		else
			t.prev = t.current
		end
	end
	ShowEffect(id)
end

local function EventCombatLog(self, ...)
	
	local args = {...}
	
	local petguid = UnitGUID("pet")
	if playerGuid ~= args[p_sourceGUID] and petguid ~= args[p_sourceGUID] then
		return
	end
	
	--print all
	--local temp = ""
	--for i = 1, #args do
	--	temp = temp..i.." = "..tostring(args[i])..", "
	--end
	--print(temp)
	
	--print partial
	if DebugSpellMode then
		local temp = tostring(args[p_event])..", "..tostring(args[p_spell_offset + p_spell_id])..", "..tostring(args[p_spell_offset + p_spell_name])
		if #args >= p_buff_offset then
			temp = temp..", "..tostring(args[p_buff_offset])
		end
		print(temp)
	end
	
	local events = {strsplit("_", args[p_event])}
	
	if events[1] == "SPELL" then
		local id = args[p_spell_offset + p_spell_id]
		local id2 = LE_SpellPartOf[id] or id
		if DebugSpellMode then
			print(id, "added to", id2)
		end
		local name = args[p_spell_offset + p_spell_name]
		
		local e = events[2]
		if e == "CAST" then
			local r = events[3]
			if r == "SUCCESS" or r == "START" then
				--InitAndShowSpell(id2)
				--ResetSpell(id2)
			elseif r == "FAILED" then
				ResetSpell(id2)
			end
		elseif e == "AURA" then
			local r = events[3]
			if r == "APPLIED" or r == "REMOVED" then
				--InitAndShowSpell(id2)
			end
		end
		
		if e == "PERIODIC" then
			e = events[3]
		end
		
		--spell_effects[id] init on click
		if not(spell_effects[id2]) then
			--some proc spell, not exists in buttons or LE_SpellPartOf
			return
		end
		
		if e == "DAMAGE" then
			--inc damage
			spell_effects[id2].damage.current = spell_effects[id2].damage.current + args[p_effect_offset + p_effect_amount]
			--print("damage", spell_effects[id2].damage.current)
		elseif e == "HEAL" then
			--inc heal
			spell_effects[id2].heal.current = spell_effects[id2].heal.current + args[p_effect_offset + p_effect_amount]
			--print("heal", spell_effects[id2].heal.current)
		end
	end
end

-- slash

local function SlashTable_debug(rest)
	DebugSpellMode = not(DebugSpellMode)
	print("DebugSpellMode = "..tostring(DebugSpellMode))
end

local function SlashTable_average(rest)
	AverageMode = not(AverageMode)
	print("AverageMode = "..tostring(AverageMode))
	LastEffect_Global.AverageMode = AverageMode
end

local SlashTable = {
	{
		['name'] = 'debug',
		['hint'] = "toggle debug mode",
		['func'] = SlashTable_debug,
		['get'] = function() return DebugSpellMode; end,
	},
	{
		['name'] = 'average',
		['hint'] = "toggle average by targets",
		['func'] = SlashTable_average,
		['get'] = function() return AverageMode; end,
	},
}

local function SlashHandler(msg, editbox)
	local command, rest = msg:match("^(%S*)%s*(.-)$")
	-- Any leading non-whitespace is captured into command
	-- the rest (minus leading whitespace) is captured into rest.
	command = string.lower(command)
	
	if command == "" or command == "help" then
		print("LastEffect slash commands:\n/le or /LastEffect, params:")
		for _, v in ipairs(SlashTable) do
			print(v.name.." : "..v.hint)
		end
	else
		local found = false
		for _, v in ipairs(SlashTable) do
			if v.name == command then
				found = true
				v.func(rest)
				break
			end
		end
		
		if not(found) then
			print("unknown command '"..command.."'")
		end
	end
end

-- main

local function ButtonOnClick(self)
	local action = ButtonToAction[self]
	if action then
		local id = ActionToId[action]
		--print(self:GetName(), "OnClick", action, id)
		if id then
			InitAndShowSpell(id)
			ResetSpell(id)
		end
	end
end

local function CheckAddDamageText(btn)
	local dt = btn.damagetext
	if not(dt) then
		btn.damagetext = btn:CreateFontString(btn:GetName() .. "DamageText", "OVERLAY", GameFontNormalSmall) --GameFontNormal
		dt = btn.damagetext
		dt:SetPoint("TOPLEFT", btn, "TOPLEFT", -5, -10)
		dt:SetPoint("TOPRIGHT", btn, "TOPRIGHT", 5, -10)
		dt:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE")
		dt:SetTextColor(1, 1, 0)
	end
end
	
local function CheckAddOnClickHook(btn)
	if not(btn.lasteffecthook) then
		btn:HookScript("OnClick", ButtonOnClick)
		btn.lasteffecthook = true
	end
end

local function ClearOldRef(id, action)
	--if id changed
	local oldId = ActionToId[action]
	if oldId ~= id and oldId then
		--hide damage text and remove reference
		--print("ClearOldRef", action, id, oldId)
		local btn = SpellIdToButton[oldId][action]
		if btn then
			if btn.damagetext then
				btn.damagetext:Hide()
			end
			SpellIdToButton[oldId][action] = nil
		end
	end
end

local function GetFullSpellName(name, rank)
	if not(rank) then
		rank = ""
	end
	if (name) then
		return name.."("..rank..")", name
	end
	return nil, nil
end

local function GetSpellId(name, rank)
	local spellName = GetFullSpellName(name, rank)
	if spellName then
		local link = GetSpellLink(spellName)
		if link then
			return tonumber(select(3, strfind(link, "spell:(%d+)|")))
		end
	end
	return nil
end

local function SetButtonRefByAction(action)
	local atype, id, _ = GetActionInfo(action)
	--print(action, ActionToButton[action]:GetName(), atype, id)
	if atype == "macro" then
		local name, rank = GetMacroSpell(id)
		id = GetSpellId(name, rank)
		--print(action, atype, name, rank, id)
	end				
	--clear old text
	--local btn = ActionToButton[action]
	--if btn and btn.damagetext then
	--	print("hiding", btn:GetName())
	--	btn.damagetext:Hide()
	--end
	
	if id and (atype == "spell" or atype == "macro") then
		--print("adding", action, atype, name, rank, id)
		SpellIdToButton[id] = SpellIdToButton[id] or {}
		SpellIdToButton[id][action] = ActionToButton[action]
		ClearOldRef(id, action)
		ActionToId[action] = id
		ShowEffect(id)
	else
		ClearOldRef(id, action)
	end
end

local function ClearTable(t)
	for k, v in pairs(t) do
		t[k] = nil
	end
end

local function RegisterFrames()
	SpellIdToButton = SpellIdToButton or {}
	ClearTable(SpellIdToButton)
	ActionToId = ActionToId or {}
	ClearTable(ActionToId)
	ActionToButton = ActionToButton or {}
	ClearTable(ActionToButton)
	ButtonToAction = ButtonToAction or {}
	ClearTable(ButtonToAction)
	--local ab1script = ActionButton1:GetScript("OnClick")
	local frame = EnumerateFrames()
	while frame do
		local fname = frame:GetName()
		if fname and frame.IsProtected and frame:IsProtected() --exists and true
		and frame.GetObjectType and frame.GetScript and fname
		and frame:GetObjectType() == "CheckButton" then
			--local script = frame:GetScript("OnClick")
			if fname:match("^BT4Button%d+$") then
				if frame.GetAction then
					CheckAddDamageText(frame)
					CheckAddOnClickHook(frame)
					local atype, action = frame:GetAction()
					ActionToButton[action] = frame
					ButtonToAction[frame] = action
					--uncomment to print all buttons spell ids
					--print(action, fname)
					SetButtonRefByAction(action)
				end			
			elseif fname:match("^ElvUI_Bar%dButton%d+$") then
				local action = SecureButton_GetModifiedAttribute(frame, "action", SecureButton_GetEffectiveButton(frame))
				if action then
					CheckAddDamageText(frame)
					CheckAddOnClickHook(frame)
					ActionToButton[action] = frame
					ButtonToAction[frame] = action
					--uncomment to print all buttons spell ids
					--print(action, fname)
					SetButtonRefByAction(action)
				end
			elseif frame.action and (frame.action < 120) then --(script == ab1script) and 
				if fname:match("^ActionButton%d+$") 
				or fname:match("^MultiBarBottomLeftButton%d+$") or fname:match("^MultiBarBottomRightButton%d+$") 
				or fname:match("^MultiBarLeftButton%d+$") or fname:match("^MultiBarRightButton%d+$") then
					CheckAddDamageText(frame)
					CheckAddOnClickHook(frame)
					local action = frame.action
					ActionToButton[action] = frame
					ButtonToAction[frame] = action
					--uncomment to print all buttons spell ids
					--local atype, id, _ = GetActionInfo(action)
					--print(action, atype, fname, id)
					SetButtonRefByAction(action)
				end
			end
		end
		frame = EnumerateFrames(frame)
	end
end

local function SlotChanged(self, ...)
	local action = ...
	--print(action, "SlotChanged") --ActionToButton[action]:GetName(), 
	SetButtonRefByAction(action)
end

local function RegisterEvents()
	EventHandlersTable = {
		["ACTIONBAR_SLOT_CHANGED"] 			= SlotChanged, 
		["ACTIONBAR_PAGE_CHANGED"]			= RegisterFrames,
		["UPDATE_SHAPESHIFT_FORM"]			= RegisterFrames,
		["COMBAT_LOG_EVENT_UNFILTERED"]		= EventCombatLog,
	}
	
	for k, _ in pairs(EventHandlersTable) do
		EventFrame:RegisterEvent(k)
	end
end

local function Init()
	LastEffect_Global = LastEffect_Global or {}
	AverageMode = LastEffect_Global.AverageMode
	
	SLASH_LASTEFFECT1, SLASH_LASTEFFECT2 = "/le", "/LastEffect"
	SlashCmdList["LASTEFFECT"] = SlashHandler
	
	playerGuid = UnitGUID("player")
	RegisterFrames() --enum buttons
	RegisterEvents()

	IsInit = true
end

EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")

local function OnEvent(self, event, ...)
	--print(event)
	if event == "PLAYER_ENTERING_WORLD" then
		EventFrame:UnregisterEvent("PLAYER_ENTERING_WORLD")
		Init()
	elseif IsInit and EventHandlersTable[event] then
		EventHandlersTable[event](self, ...)
	end
end

EventFrame:SetScript("OnEvent", OnEvent)
